#!/usr/bin/env python

# superchick - view manga (or other collections of images)
# Copyright (c)2003 Joe Wreschnig <wres0003@umn.edu>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
                                                                               
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os, sys
import popen2
import getopt

__author__    = "Joe Wreschnig <wres0003@umn.edu>"
__copyright__ = "Copyright (c)2003 Joe Wreschnig <wres0003@umn.edu>"
__revision__  = "$Id$"
__version__   = "0.4"

VERSION_TEXT = """\
Superchick %s, by %s
http://kai.vm.bytemark.co.uk/~piman/software.shtml

This is free software; you are welcome to redistribute it under the
terms of the GNU GPL version 2 as published by the Free Software
Foundation; a copy of the GPL v2 should have been included with this
program as the file `COPYING'.""" %(__version__, __author__)


SHORT_USAGE = "Usage: %s [options] archive-or-directory" % sys.argv[0]
USAGE = """\
%s

Options:
  --ltr, -l         default to left-to-right mode
  --rtl, -r         default to right-to-left mode
  --prompt, -p      prompt for a filename if none is given
""" % SHORT_USAGE

BLACK = [0, 0, 0]
GRAY = [200, 200, 200]
WHITE = [255, 255, 255]

# Image file formats matched in directories and zips.
EXTENSIONS = [".jpg", "jpeg", ".gif", ".png", ".bmp", ".pcx", ".tif", "tiff"]

def has_unrar():
    for path in os.environ["PATH"].split(os.pathsep):
        if os.path.isfile(os.path.join(path, 'unrar')):
            return True
    return False

# Scale an image maintaining its aspect ratio. Use rotozoom so we get
# a filtered image.
def aspect_scale(image, (bx, by)):
    ix, iy = image.get_size()
    scale_factor = min(1, bx / float(ix), by / float(iy))
    return pygame.transform.rotozoom(image, 0, scale_factor)

class Config(dict):
    def load(self, filename):
        f = file(filename)
        for line in f:
            fn, marks = line.split("\t")
            self[fn] = map(int, marks.split())

    def write(self, filename):
        f = file(filename, "w")
        for key, val in self.items():
            f.write("%s\t%s\n" % (key, " ".join(map(str, val))))

class Store(object):
    def __len__(self):
        return len(self._files)

# Thanks to Bill Witherspoon for the base implementation of this class.
class RarFile(object):
    def __init__(self, filename, mode):
        thefiles = os.popen2(["unrar", "vb", "-c-", filename])[1].readlines()
        self._files = [name.strip() for name in thefiles]
        self.filename = filename

    def namelist(self):
        return self._files

    def read(self, name):
        f = os.popen2(["unrar","p", "-c-", "-inul", self.filename, name])[1]
        return f.read()

# A set of files in a directory.
class DirStore(Store):
    def __init__(self, filename):
        self._filename = os.path.realpath(filename)
        self._files = [os.path.join(filename, i) for i in os.listdir(filename)
                       if i[-4:].lower() in EXTENSIONS]
        self._files.sort()

    def __getitem__(self, i): return self._files[i]

# A set of files in a zip file.
class ZipStore(Store):
    def __init__(self, filename):
        self._filename = os.path.realpath(filename)
        if (filename.lower().endswith(".zip") or
            filename.lower().endswith(".cbz")):
            self._zip = zipfile.ZipFile(filename, "r")
        elif (has_unrar() and
              (filename.lower().endswith(".rar") or
               filename.lower().endswith(".cbr"))) :
            self._zip = RarFile(filename, "r")
        elif (filename.lower().endswith(".tar.gz") or
              filename.lower().endswith(".tgz") or
              filename.lower().endswith(".tar.bz2") or
              filename.lower().endswith(".tb2") or
              filename.lower().endswith(".tbz")):
            self._zip = tarfile.TarFileCompat(filename, "r")
        else: raise ValueError("unknown file type based on extension")
        self._files = [i for i in self._zip.namelist()
                       if i[-4:].lower() in EXTENSIONS]
        self._files.sort()

    def __getitem__(self, i):
        return StringIO(self._zip.read(self._files[i]))

# A cache object. Although we can't actually speed up image load, we can
# store recently loaded (and images we expect to need soon) precached.
# However, manga images are huge, so we can't have too many.
class ImageCache(object):
    def __init__(self, items):
        self.items = items
        self.cached = []
        self.cache = {}

    def clear(self):
        self.cached = []
        self.cache = {}

    def __setitem__(self, key, val):
        if key not in self.cached:
            self.cached.append(key)
            if len(self.cached) > self.items:
                to_del = self.cached[0]
                self.cached.pop(0)
                del(self.cache[to_del])
        self.cache[key] = val

    def __getitem__(self, key):
        return self.cache.get(key)

class Manga(object):
    def __init__(self, filename, config, rtl = True):
        if os.path.isdir(filename): self._store = DirStore(filename)
        else: self._store = ZipStore(filename)

        self.filename = filename
        self._image = self._image2 = None
        self._config = config
        self._bookmarks = self._config.get(filename, [-1] * 10)

        if len(self._bookmarks) == 11:
            self.rtl = bool(self._bookmarks[-1])
        else:
            self.rtl = rtl
            self._bookmarks.append(int(self.rtl))

        self._config[self.filename] = self._bookmarks
        self._cache = ImageCache(10)
        self._page = self._bookmarks[0]
        if self._page == -1: self._page = 0

    def toggle_rtl(self):
        self.rtl ^= True
        self._bookmarks[-1] = int(self.rtl)

    # Call this when the manga is first read, to get the first pages.
    # Don't do it as soon as we instantiate the manga, because the
    # Store might be empty.
    def opened(self): self._load_page()

    # Set a bookmark.
    def set_mark(self, markno):
        self._bookmarks[markno] = self._page

    # Go to a bookmark and load the pages.
    def go_to_mark(self, markno):
        if self._bookmarks[markno] != -1:
            self._page = self._bookmarks[markno]
            self._load_page()

    def __len__(self):
        return len(self._store)

    def page(self): return self._page + 1

    # Skip forward two images if we're displaying two at once.
    def next_screen(self):
        if self._image2: self.next(2)
        else: self.next()

    def go_to(self, page):
        self._page = min(page, len(self) - 1)
        self._load_page()

    def next(self, count = 1):
        self._page = min(self._page + count, len(self) - 1)
        self._load_page()

    def previous(self, count = 1):
        self._page = max(self._page - count, 0)
        self._load_page()

    # Load an image and render it; if there's space left, load another.
    def _load_page(self):
        self._image = self._cache[self._page]
        if self._image is None:
            self._image = pygame.image.load(self._store[self._page])
            self._image = aspect_scale(self._image, [self._w, self._h])
            self._cache[self._page] = self._image

        if (self._w / 4 <= self._image.get_width() <= self._w / 2 and
            self._page != len(self) - 1):
            self._image2 = self._cache[self._page + 1]
            if self._image2 is None:
                self._image2 = pygame.image.load(self._store[self._page + 1])
                self._image2 = aspect_scale(self._image2, [self._w, self._h])
                self._cache[self._page + 1] = self._image2

            if (self._image2.get_width() > self._w / 2 or
                self._image2.get_width() < self._w / 4): self._image2 = None
        else: self._image2 = None

    # This behavior is important to note - Scrolling through images takes
    # just as long with the cache as without the cache. However in the
    # common case of "display an image, read it, repeat", this makes
    # the to-be-read image display right away as the next one loads
    # in the background.
    def preload(self):
        for i in [-1, 0, 1, 2, 3]:
            if (len(self) > self._page + i >= 0 and
                not self._cache[self._page + i]):
                img = pygame.image.load(self._store[self._page + i])
                self._cache[self._page + i] = aspect_scale(img,
                                                           [self._w, self._h])

    def set_size(self, size):
        self._w = size[0]
        self._h = int(self._w / (1024.0 / 730.0))
        self._cache.clear()
        self._load_page()

    def render(self):
        surf = pygame.Surface([self._w, self._h])
        surf.fill(BLACK)

        if self._image2:
            r1 = self._image.get_rect()
            r1.centery = surf.get_rect().centery
            r2 = self._image2.get_rect()
            r2.centery = surf.get_rect().centery
            if self.rtl:
                r1.left = surf.get_rect().centerx + 1
                surf.blit(self._image, r1)
                r2.right = surf.get_rect().centerx - 1
                surf.blit(self._image2, r2)
            else:
                r2.right = surf.get_rect().centerx + 1
                surf.blit(self._image, r2)
                r1.left = surf.get_rect().centerx - 1
                surf.blit(self._image2, r1)

        else:
            r = self._image.get_rect()
            r.center = surf.get_rect().center
            surf.blit(self._image, r)

        return surf

# A text box class with tab completion.
class TextBox(object):
    def __init__(self, message, tabcomp = True, size = 28, width = 512):
        if message == None: message = ""

        font = pygame.font.Font(None, size)
        surf = pygame.Surface([width + 2, 2 * font.get_linesize() + 4])
        font.set_bold(True)
        question = font.render(message, True, GRAY)
        surf.fill(BLACK)
        pygame.draw.rect(surf, WHITE, surf.get_rect(), 1)
        surf.blit(question, [1, 1])
        font.set_bold(False)

        self._tabcomp = tabcomp
        self._font = font
        self._surface = surf

    def ask(self, screen, string = None):
        if string == None: string = ""
        idx = len(string)
        cut = ""

        r_q = self._surface.get_rect()
        r_q.center = screen.get_rect().center
        while True:
            x_off = 10
            screen.blit(self._surface, r_q)
            for i in range(len(string)):
                c = string[i]
                if i == idx: self._font.set_underline(True)                    
                img_a = self._font.render(c, True, WHITE)
                self._font.set_underline(False)
                r_a = img_a.get_rect()
                r_a.topleft = [r_q.x + x_off,
                               r_q.y + self._font.get_linesize() + 1]
                x_off += img_a.get_width()            
                screen.blit(img_a, r_a)

            if idx == len(string):
                img_a = self._font.render("_", True, WHITE)
                r_a = img_a.get_rect()
                r_a.topleft = [r_q.x + x_off,
                               r_q.y + self._font.get_linesize() + 1]
                screen.blit(img_a, r_a)

            pygame.display.update()

            ev = pygame.event.wait()
            while ev.type != KEYDOWN:
                ev = pygame.event.wait()

            if ev.key == K_ESCAPE: return None
            elif ev.key == K_F11: pygame.display.toggle_fullscreen()
            elif ev.key == K_BACKSPACE:
                string = string[:idx - 1] + string[idx:]
                idx -= 1
            elif ev.key == K_DELETE:
                string = string[:idx] + string[idx + 1:]
            elif ev.key == K_RETURN: break
            elif ev.key == K_END: idx = len(string)
            elif ev.key == K_HOME: idx = 0
            elif ev.key == K_LEFT: idx = max(0, idx - 1)
            elif ev.key == K_RIGHT: idx = min(len(string), idx + 1)
            elif ev.key == K_TAB:
                if self._tabcomp: string = self._complete(string)
                idx = len(string)
            elif ev.key < 128:
                if ev.mod & KMOD_CTRL:
                    if len(string) == 0: pass
                    elif ev.key == K_u:
                        string = ""
                        idx = 0
                    elif ev.key == K_h: string = string[0:-1]
                    elif ev.key == K_e: idx = len(string)
                    elif ev.key == K_a: idx = 0
                    elif ev.key == K_f: idx = min(len(string), idx + 1)
                    elif ev.key == K_b: idx = max(0, idx - 1)
                    elif ev.key == K_t:
                        if idx == 0: idx = 1
                        if len(string) == 1: pass
                        elif idx == len(string):
                            string = string[:-2] + string[-1] + string[-2]
                        else:
                            string = (string[:idx - 1] + string[idx] +
                                      string[idx - 1] + string[idx + 1:])
                            idx = min(len(string), idx + 1)
                    elif ev.key == K_k:
                        cut = string[idx:]
                        string = string[:idx]
                    elif ev.key == K_y:
                        string = string[:idx] + cut + string[idx:]
                        idx += len(cut)
                    elif ev.key == K_w:
                        if string[-1] == "/": string = string[:-1]
                        string = string[:string.rfind("/") + 1]
                else:
                    string = string[:idx] + ev.unicode + string[idx:]
                    idx += 1

        return string

    def _complete(self, string):
        if '*' not in string: string += "*"

        g = glob.glob(string)
        if len(g) == 0: s = string[:-1]
        elif len(g) == 1: s = g[0]
        else: # thanks, mu
            for i, chars in enumerate(zip(*g)):
                if min(chars) != max(chars):
                    j = i
                    break
            else: j = min([len(s) for s in g])
            s = g[0][:j]
        s = os.path.expanduser(s)
        if os.path.isdir(s) and s[-1] != "/": s += "/"
        return s

class ReaderUI(object):
    def __init__(self, screen, rtl, initial_fn = None):
        self._screen = screen
        self._sfont = pygame.font.Font(None, 14)
        self._rtl = rtl
        self._size = 0

        self._manga = self._get_manga(initial_fn)
        if not self._manga: return

        self._manga.set_size(ReaderUI.sizes[self._size])
        self._manga.opened()

        self.render()

        while True:
            ev = pygame.event.wait()

            if ev.type == QUIT: break
            elif ev.type == MOUSEBUTTONDOWN:
                if ev.button == 1:
                    if ev.pos >= self._screen.get_size()[0]/2:
                        if self._manga.rtl: self._manga.previous()
                        else: self._manga.next()
                    else:
                        if self._manga.rtl: self._manga.next()
                        else: self._manga.previous()
                elif ev.button == 3:
                    self._manga = self._get_manga(self._manga.filename)
                    if not self._manga: return
                    self._manga.set_size(ReaderUI.sizes[self._size])
                    self._manga.opened()
                elif ev.button == 4: self._manga.previous()
                elif ev.button == 5: self._manga.next()
            elif ev.key == K_MINUS:
                self._size = (self._size + 1) % len(ReaderUI.sizes)
                pygame.display.set_mode(ReaderUI.sizes[self._size])
                self._manga.set_size(ReaderUI.sizes[self._size])
            elif ev.key == K_EQUALS:
                self._size = (self._size - 1) % len(ReaderUI.sizes)
                pygame.display.set_mode(ReaderUI.sizes[self._size])
                self._manga.set_size(ReaderUI.sizes[self._size])
            elif ev.key == K_ESCAPE: break
            elif ev.key == K_SPACE: self._manga.next_screen()
            elif ev.key == K_RIGHT:
                if self._manga.rtl: self._manga.previous()
                else: self._manga.next()
            elif ev.key == K_LEFT:
                if self._manga.rtl: self._manga.next()
                else: self._manga.previous()
            elif ev.key == K_PAGEUP: self._manga.previous(15)
            elif ev.key == K_PAGEDOWN: self._manga.next(15)
            elif ev.key in [K_f, K_F11]: pygame.display.toggle_fullscreen()
            elif ev.key == K_r: self._manga.toggle_rtl()
            elif ev.key == K_HOME: self._manga.go_to(0)
            elif ev.key == K_END: self._manga.go_to(len(self._manga))
            elif ev.key == K_l:
                self._manga = self._get_manga(self._manga.filename)
                if not self._manga: return
                self._manga.opened()
            elif K_0 <= ev.key <= K_9:
                if ev.mod & KMOD_SHIFT:
                    self._manga.set_mark(ev.key - K_0)
                else: self._manga.go_to_mark(ev.key - K_0)

            # Bookmark 0 is the current / most recent page.
            self._manga.set_mark(0)
            self.render()

    def _get_manga(self, filename):
        box = TextBox("Enter a filename (zip or directory of images):")
        while True:
            new_fn = box.ask(self._screen, filename)

            if new_fn is not None:
                if new_fn == "":
                    if filename: return Manga(filename, config, self._rtl)
                elif not os.path.exists(new_fn):
                    box = TextBox("File not found. Enter a filename:")
                    filename = new_fn
                else:
                    try: m = Manga(new_fn, config, self._rtl)
                    except ValueError: m = []

                    if len(m) == 0:
                        box = TextBox("Unable to open. Enter a filename:")
                        filename = new_fn
                    else: return m
            else: return None

    def render(self):
        self._screen.fill(BLACK)
        img = self._manga.render()
        r = img.get_rect()
        r.center = self._screen.get_rect().center
        self._screen.blit(img, r)

        top_text = "Reading: %s" % self._manga.filename
        if self._manga.rtl: top_text += " (RTL)"
        else: top_text += " (LTR)"

        bottom_text = "Page: %d/%d" % (self._manga.page(), len(self._manga))

        t1 = self._sfont.render(top_text, True, WHITE, BLACK)
        t2 = self._sfont.render(bottom_text, True, WHITE, BLACK)

        self._screen.blit(t1, [0, 0])
        self._screen.blit(t2, [0, 754])
            
        pygame.display.update()

        self._manga.preload()

opts, args = getopt.getopt(sys.argv[1:], "vhlrp",
                           ["help", "version", "ltr", "rtl", "prompt"])

rtl = True
prompt = False

for o, a in opts:
    if o in ["-v", "--version"]:
        print VERSION_TEXT
        raise SystemExit
    elif o in ["-h", "--help"]:
        print USAGE
        raise SystemExit
    elif o in ["-l", "--ltr"]: rtl = False
    elif o in ["-r", "--rtl"]: rtl = True
    elif o in ["-p", "--prompt"]: prompt = True

if len(args) != 1:
    if not prompt:
        print SHORT_USAGE
        print "(Or try %s --help)" % sys.argv[0]
        raise SystemExit(1)
    else: filename = None
else: filename = os.path.realpath(args[0])

if os.name == "posix": config_fn = "%s/.superchickrc" % os.environ["HOME"]
else: config_fn = "superchickrc"

import glob
import string
import zipfile
import tarfile
from cStringIO import StringIO

import pygame
from pygame.locals import *

config = Config()
if os.path.exists(config_fn): config.load(config_fn)

try:
    import psyco
    print "Psyco optimizing compiler found."
    psyco.full()
except ImportError: pass

pygame.display.init()
pygame.font.init()

pygame.event.set_blocked(range(NUMEVENTS))
pygame.event.set_allowed([KEYDOWN, MOUSEBUTTONDOWN, QUIT])
pygame.key.set_repeat(500, 30)

ReaderUI.sizes = pygame.display.list_modes(0, FULLSCREEN)
screen = pygame.display.set_mode(ReaderUI.sizes[0], FULLSCREEN)
screen.fill(BLACK)
ReaderUI(screen, rtl, filename)

config.write(config_fn)
